#!/usr/bin/env python

"""
View annotated sources.

Copyright (C) 2006, 2007 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from compiler.visitor import ASTVisitor
from simplify.simplified import *
import sys
import os
import textwrap

# Classes.

# HTML-related output production.

html_header = """<?xml version="1.0" encoding="iso-8859-15"?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml">
<head>
  <title>Module</title>
  <style type="text/css">
    html {
      background-color: black; color: white;
    }

    body {
      padding-bottom: 4em;
      font-size: 14pt; font-family: monospace;
      background-color: black; color: white;
    }

    .class { margin-top: 1em; margin-bottom: 1em; }
    .function { margin-top: 1em; margin-bottom: 1em; }
    .body { padding-left: 2em; }
    .keyword { color: yellow; }
    .comment { color: blue; }
    .class-name { color: cyan; }
    .function-name { color: cyan; }
    .str { color: #FF00FF; }
    .doc { color: #FF00FF; margin-top: 1em; margin-bottom: 1em; }
    .invocation a { color: white; text-decoration: none; }

    .popup {
      display: none; z-index: 2;
      position: absolute; top: 2ex; left: 0;
      padding: 0.2em; background-color: #000000; color: white;
      border: 2px solid #dddddd;
    }

    .invocations {
      padding: 0.5em; background-color: #770000;
      clear: all;
    }

    .types {
      padding: 0.5em; background-color: #0000FF;
      float: right;
    }

    .raises {
      padding: 0.5em; background-color: #7700FF;
      float: right;
    }

    .scopes {
      padding: 0.5em; background-color: #007700;
      float: left;
    }

    .non-writes, .non-accesses {
      padding: 0.5em; background-color: #FF0000;
      float: right;
    }

    .no-types {
      background-color: #FF0000;
    }

    .op,
    .name,
    .attr,
    .conditional,
    .operator,
    .iterator,
    .call,
    .returns,
    .failure
    {
      position: relative;
    }

    .op:hover > .popup,
    .name:hover > .popup,
    .attr:hover > .popup,
    .conditional:hover > .popup,
    .operator:hover > .popup,
    .iterator:hover > .popup,
    .call:hover > .popup,
    .returns:hover > .popup,
    .failure:hover > .popup
    {
       display: block;
    }

    .summary-class {
      background-color: #004400;
    }

    .summary-instance {
      background-color: #0000FF;
    }

    .summary-attr {
      background-color: #007700;
    }

  </style>
</head>
<body>
"""

html_footer = """</body>
</html>
"""

# Utility classes.

class Writer:

    "A utility class providing useful HTML output methods."

    # Methods which return strings.

    def _text(self, text):
        return text.replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")

    def _attr(self, attr):
        return self._text(attr).replace("'", "&apos;").replace('"', "&quot;")

    def _url(self, url):
        return self._attr(url).replace("#", "%23").replace("-", "%2d")

    def _summary_link(self, module_name, name):
        return "<a href='%s-summary%sxhtml#%s'>%s</a>" % (module_name, os.path.extsep, self._attr(name), self._text(name))

    # Methods which write to the stream.

    def _comment(self, comment):
        self.stream.write("<span class='comment'># %s</span>\n" % comment)

    def _keyword(self, kw):
        self.stream.write("<span class='keyword'>%s</span> " % kw)

    def _doc(self, node):
        if node.doc is not None:
            self.stream.write("<pre class='doc'>\n")
            self.stream.write('"""')
            output = textwrap.dedent(node.doc.replace('"""', '\\"\\"\\"'))
            self.stream.write(self._text(output))
            self.stream.write('"""')
            self.stream.write("</pre>\n")

    def _name(self, name):
        self.stream.write("<span class='name'>%s</span>\n" % name)

# Summary classes.

class Summariser(Writer):

    def __init__(self, stream):
        self.stream = stream

    def process(self, module):
        self.module = module
        self._init_details()
        self.stream.write(html_header)
        self._write_classes(module)
        self.stream.write(html_footer)

    def _write_classes(self, module):
        self.stream.write("<table cellspacing='5' cellpadding='5'>\n")
        for structure in self.structures:
            self._write_class(structure)
        self.stream.write("</table>\n")

    def _write_class(self, structure):

        # Write the class...

        self.stream.write("<tbody class='class'>\n")
        self.stream.write("<tr>\n")
        self.stream.write("<th class='summary-class' id='%s'>\n" % structure.full_name())
        self._keyword("class")
        self.stream.write(structure.name)
        self.stream.write("</th>\n")

        # ...and all known attribute names.

        structure_attributes = self.structure_attributes[structure]
        for name in self.attribute_names:
            if name in structure_attributes:
                self.stream.write("<th class='summary-attr'>%s</th>\n" % self._text(name))
            else:
                self.stream.write("<th></th>\n")
        self.stream.write("</tr>\n")

        # Write instances for the class, along with type details for each attribute.

        for instance in self.structures_to_instances[structure]:
            self.stream.write("<tr>\n")
            self.stream.write("<th class='summary-instance'>%s</th>\n" % self._text(instance.full_name()))
            for name in self.attribute_names:
                self.stream.write("<td>\n")
                values = instance.namespace.get(name, [])
                type_names = [value.type.full_name() for value in values]
                type_names.sort()
                for type_name in type_names:
                    self.stream.write("<span>%s</span>\n" % type_name)
                self.stream.write("</td>\n")
            self.stream.write("</tr>\n")
        self.stream.write("</tbody>\n")

    def _init_details(self):
        names = set()
        self.structures_to_instances = {}
        self.structure_attributes = {}
        names_to_structures = {}

        # Visit all structures.

        for structure in self.module.simplifier.structures:

            # Map all instance names to instances.

            names_to_instances = {}
            attribute_names = set()

            for instance in structure.instances.values():
                for name, values in instance.namespace.items():
                    if not name in attribute_names:
                        attribute_names.add(name)
                        if not name in names:
                            names.add(name)
                names_to_instances[instance.full_name()] = instance

            # Record the instances in name order for the structure.

            instance_names = names_to_instances.keys()
            instance_names.sort()
            sorted_instances = [names_to_instances[name] for name in instance_names]
            self.structures_to_instances[structure] = sorted_instances

            # Record the attributes used in all instances of the structure.

            self.structure_attributes[structure] = attribute_names

            # Record the name to structure mapping.

            names_to_structures[structure.full_name()] = structure

        self.attribute_names = list(names)
        self.attribute_names.sort()

        structure_names = names_to_structures.keys()
        structure_names.sort()

        self.structures = [names_to_structures[name] for name in structure_names]

# Browser classes.

class Browser(ASTVisitor, Writer):

    """
    A browsing visitor for AST nodes.

    Covered: Add, And, Assert, AssAttr, AssList, AssName, AssTuple, Assign,
             AugAssign, Bitand, Break, CallFunc, Class, Compare, Const,
             Continue, Dict, Discard, Div, FloorDiv, For, From, Function,
             Getattr, Global, If, Import, Keyword, Lambda, List, ListComp,
             ListCompFor, ListCompIf, Mod, Module, Mul, Name, Not, Or, Pass,
             Power, Print, Printnl, Raise, Return, Slice, Sliceobj, Stmt, Sub,
             Subscript, TryExcept, TryFinally, Tuple, UnaryAdd, UnarySub, While.

    Missing: Backquote, Bitor, Bitxor, Decorators, Ellipsis,
             Exec, Invert, LeftShift, RightShift, Yield.
    """

    def __init__(self, stream):
        ASTVisitor.__init__(self)
        self.visitor = self
        self.stream = stream

    def process(self, module):
        self.module = module
        self.stream.write(html_header)
        self.dispatch(module)
        self.stream.write(html_footer)

    def visitModule(self, node):
        self.default(node)

    # Statements.

    def visitAssert(self, node):
        self.stream.write("<div class='assert'>\n")
        self.stream.write("<span class='failure'>\n")
        self._keyword("assert")
        self._popup(
            self._types(node._raises.active())
            )
        self.stream.write("</span>\n")
        self.dispatch(node.test)
        if node.fail:
            self.stream.write(", ")
            self.dispatch(node.fail)
        self.stream.write("</div>\n")

    def visitAssign(self, node):
        self.stream.write("<div class='assign'>\n")
        for lvalue in node.nodes:
            self.dispatch(lvalue)
            self.stream.write("=\n")
        self.dispatch(node.expr)
        self.stream.write("</div>\n")

    def visitAugAssign(self, node):
        self.stream.write("<div class='augassign'>\n")
        self.dispatch(node.node)
        self.stream.write("<span class='operator'>\n")
        self.stream.write("%s\n" % node.op)
        self._popup(
            self._invocations(node._op_call.active())
            )
        self.stream.write("</span>\n")
        self.dispatch(node.expr)
        self.stream.write("</div>\n")

    def visitBreak(self, node):
        self.stream.write("<div class='break'>\n")
        self._keyword("break")
        self.stream.write("</div>\n")

    def visitClass(self, node):
        definition = node._node
        definitions = definition.active()
        structure = definition.expr.ref
        self.stream.write("<div class='class' id='%s'>\n" % structure.full_name())
        self.stream.write("<div>\n")
        self._keyword("class")
        self._name_start(structure.name, "class-name")
        self._popup(
            self._scopes(definitions)
            )
        self._name_end()
        bases = structure.bases

        # Suppress the "object" class appearing alone.

        if bases and not (len(bases) == 1 and bases[0].name == "object"):
            self.stream.write("(")
            first = 1
            for base in bases:
                if not first:
                    self.stream.write(",\n")
                self._name_start(base.name)
                self._popup(
                    self._scopes([base]) +
                    self._types([base])
                    )
                self._name_end()
                first = 0
            self.stream.write(")")

        self.stream.write(":\n")
        self._comment(self._summary_link(self.module._node.name, structure.full_name()))
        self.stream.write("</div>\n")

        self.stream.write("<div class='body'>\n")
        self._doc(node)
        self.dispatch(node.code)
        self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitContinue(self, node):
        self.stream.write("<div class='continue'>\n")
        self._keyword("continue")
        self.stream.write("</div>\n")

    def visitDiscard(self, node):
        self.stream.write("<div class='discard'>\n")
        self.default(node)
        self.stream.write("</div>\n")

    def visitFor(self, node):
        self.stream.write("<div class='if'>\n")
        self.stream.write("<div>\n")
        self.stream.write("<span class='iterator'>\n")
        self._keyword("for")
        self._popup(
            self._invocations(node._next_call.active())
            )
        self.stream.write("</span>\n")
        self.dispatch(node.assign)
        self.stream.write("<span class='iterator'>\n")
        self._keyword("in")
        self._popup(
            self._invocations(node._iter_call.active())
            )
        self.stream.write("</span>\n")
        self.dispatch(node.list)
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body'>\n")
        self.dispatch(node.body)
        self.stream.write("</div>\n")
        if node.else_ is not None:
            self.stream.write("<div>\n")
            self._keyword("else")
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body'>\n")
            self.dispatch(node.else_)
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitFrom(self, node):
        self.stream.write("<div class='from'>\n")
        self._keyword("from")
        self.stream.write("<span class='name'>\n")
        self.stream.write(node.modname)
        self._popup(
            self._types(node._modname.active())
            )
        self.stream.write("</span>\n")
        self._keyword("import")
        first = 1
        for (name, alias), _name in map(None, node.names, node._names):
            if not first:
                self.stream.write(",\n")
            if alias:
                self.stream.write(name + " ")
                self._keyword("as")
            self.stream.write("<span class='name'>\n")
            self.stream.write(alias or name)
            self._popup(
                self._types([_name])
                )
            self.stream.write("</span>\n")
            first = 0
        self.stream.write("</div>\n")

    def visitFunction(self, node):
        definition = node._node
        definitions = [n for n in definition.active() if not isinstance(n, Subprogram)]
        subprogram = node._subprogram
        subprograms = subprogram.active()
        self.stream.write("<div class='function' id='%s'>\n" % subprogram.full_name())
        self.stream.write("<div>\n")
        self._keyword("def")
        self._name_start(subprogram.name, "function-name")
        self._popup(
            self._scopes([definition]) + # not dependent on subprograms
            self._raises(subprograms)
            )
        self._name_end()
        self.stream.write("(")
        self._parameters(subprogram, subprograms)
        self.stream.write(")")
        self.stream.write(":\n")
        self._comment(self._text(subprogram.full_name()))
        self.stream.write("</div>\n")

        self.stream.write("<div class='body'>\n")
        self._doc(node)
        self.dispatch(node.code)
        self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitGlobal(self, node):
        self.stream.write("<div class='global'>\n")
        self._keyword("global")
        first = 1
        for name in node.names:
            if not first:
                self.stream.write(",\n")
            self.stream.write(name)
            first = 0
        self.stream.write("</div>\n")

    def visitIf(self, node):
        self.stream.write("<div class='if'>\n")
        first = 1
        conditional = node._node
        conditionals = conditional.active()
        for compare, stmt in node.tests:
            self.stream.write("<div>\n")
            self.stream.write("<span class='conditional'>\n")
            if first:
                self._keyword("if")
            else:
                self._keyword("elif")
            self._popup(
                self._invocations([c.test for c in conditionals])
                )
            self.stream.write("</span>\n")
            self.dispatch(compare)
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body'>\n")
            self.dispatch(stmt)
            self.stream.write("</div>\n")
            if conditional.else_:
                conditional = conditional.else_[0]
                conditionals = conditional.active()
            else:
                conditional = None
                conditionals = []
            first = 0
        if node.else_ is not None:
            self.stream.write("<div>\n")
            self._keyword("else")
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body'>\n")
            self.dispatch(node.else_)
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitImport(self, node):
        self.stream.write("<div class='import'>\n")
        self._keyword("import")
        first = 1
        for (name, alias), _name in map(None, node.names, node._names):
            if not first:
                self.stream.write(",\n")
            if alias:
                self.stream.write(name + " ")
                self._keyword("as")
            self.stream.write("<span class='name'>\n")
            self.stream.write(alias or name)
            self._popup(
                self._types([_name])
                )
            self.stream.write("</span>\n")
            first = 0
        self.stream.write("</div>\n")

    def visitPass(self, node):
        self.stream.write("<div class='pass'>\n")
        self._keyword("pass")
        self.stream.write("</div>\n")

    def visitPrint(self, node):
        self.stream.write("<div class='print'>\n")
        self._keyword("print")
        if node.dest is not None:
            self.stream.write(">>\n")
            self.dispatch(node.dest)
        for n in node.nodes:
            self.dispatch(n)
            self.stream.write(",\n")
        self.stream.write("</div>\n")

    def visitPrintnl(self, node):
        self.stream.write("<div class='printnl'>\n")
        self._keyword("print")
        if node.dest is not None:
            self.stream.write(">>\n")
            self.dispatch(node.dest)
        first = 1
        for n in node.nodes:
            if not first:
                self.stream.write(",\n")
            self.dispatch(n)
            first = 0
        self.stream.write("</div>\n")

    def visitRaise(self, node):
        target = node._node.expr
        targets = target.active()
        self.stream.write("<div class='raise'>\n")
        self.stream.write("<span class='call'>\n")
        self._keyword("raise")
        self._popup(
            self._invocations(targets)
            )
        self.stream.write("</span>\n")
        self.dispatch(node.expr1)
        if node.expr2 is not None:
            self.stream.write(",\n")
            self.dispatch(node.expr2)
        if node.expr3 is not None:
            self.stream.write(",\n")
            self.dispatch(node.expr3)
        self.stream.write("</div>\n")

    def visitReturn(self, node):
        value = node._node
        values = value.active()
        self.stream.write("<div class='return'>\n")
        self.stream.write("<span class='returns'>\n")
        self._keyword("return")
        self._popup(
            self._types(values)
            )
        self.stream.write("</span>\n")
        self.dispatch(node.value)
        self.stream.write("</div>\n")

    def visitStmt(self, node):
        self.stream.write("<div class='stmt'>\n")
        self.default(node)
        self.stream.write("</div>\n")

    def visitTryExcept(self, node):
        self.stream.write("<div class='tryexcept'>\n")
        self.stream.write("<div>\n")
        self._keyword("try")
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body'>\n")
        self.dispatch(node.body)
        self.stream.write("</div>\n")
        for spec, assign, statement in node.handlers:
            self.stream.write("<div>\n")
            self._keyword("except")
            if spec is not None:
                self.dispatch(spec)
            if assign is not None:
                self.stream.write(",\n")
                self.dispatch(assign)
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body'>\n")
            self.dispatch(statement)
            self.stream.write("</div>\n")
        if node.else_ is not None:
            self.stream.write("<div>\n")
            self._keyword("else")
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body'>\n")
            self.dispatch(node.else_)
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitTryFinally(self, node):
        self.stream.write("<div class='tryfinally'>\n")
        self.stream.write("<div>\n")
        self._keyword("try")
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body'>\n")
        self.dispatch(node.body)
        self.stream.write("</div>\n")
        self.stream.write("<div>\n")
        self._keyword("finally")
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body'>\n")
        self.dispatch(node.final)
        self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    def visitWhile(self, node):
        self.stream.write("<div class='while'>\n")
        self.stream.write("<div>\n")
        self.stream.write("<span class='conditional'>\n")
        self._keyword("while")
        self._popup(
            self._invocations(node._test_call.active())
            )
        self.stream.write("</span>\n")
        self.dispatch(node.test)
        self.stream.write(":\n")
        self.stream.write("</div>\n")
        self.stream.write("<div class='body'>\n")
        self.dispatch(node.body)
        self.stream.write("</div>\n")
        if node.else_ is not None:
            self.stream.write("<div>\n")
            self._keyword("else")
            self.stream.write(":\n")
            self.stream.write("</div>\n")
            self.stream.write("<div class='body'>\n")
            self.dispatch(node.else_)
            self.stream.write("</div>\n")
        self.stream.write("</div>\n")

    # Expression-related helper methods.

    def _visitBinary(self, node, name, symbol):
        self.stream.write("<span class='%s'>\n" % name)
        self.dispatch(node.left)
        self.stream.write("<span class='operator'>\n")
        self.stream.write(self._text(symbol))
        self._popup(
            self._invocations(node._left_call.active() + node._right_call.active())
            )
        self.stream.write("</span>\n")
        self.dispatch(node.right)
        self.stream.write("</span>")

    def _visitUnary(self, node, name, symbol):
        self.stream.write("<span class='%s'>\n" % name)
        self.stream.write("<span class='operator'>\n")
        self.stream.write(symbol)
        self._popup(
            self._invocations(node._unary_call.active())
            )
        self.stream.write("</span>\n")
        self.dispatch(node.expr)
        self.stream.write("</span>")

    # Expressions.

    def visitAdd(self, node):
        self._visitBinary(node, "add", "+")

    def visitAnd(self, node):
        self.stream.write("<span class='and'>\n")
        first = 1
        for n in node.nodes:
            if not first:
                self._keyword("and")
            self.dispatch(n)
            first = 0
        self.stream.write("</span>")

    def visitAssAttr(self, node):
        target = node._node
        targets = target.active()
        self.stream.write("<span class='assattr'>\n")
        self.dispatch(node.expr)
        self.stream.write("<span class='attr'>\n")
        self.stream.write(".")
        types = self._types(targets)
        if not target.is_annotated() or types:
            self._name_start(node.attrname)
        else:
            self._name_start(node.attrname, "no-types")
        self._popup(
            self._scopes(targets) +
            types
            )
        self._name_end()
        self.stream.write("</span>\n")
        self.stream.write("</span>\n")

    def visitAssList(self, node):
        self.stream.write("<span class='list'>\n")
        self.stream.write("[")
        self._sequence(node)
        self.stream.write("]\n")
        self.stream.write("</span>\n")

    def visitAssName(self, node):
        target = node._node
        targets = target.active()
        types = self._types(targets)
        if not target.is_annotated() or types:
            self._name_start(target.name)
        else:
            self._name_start(target.name, "no-types")
        self._popup(
            self._scopes(targets) +
            types
            )
        self._name_end()

    def visitAssTuple(self, node):
        self.stream.write("<span class='tuple'>\n")
        self.stream.write("(")
        self._sequence(node)
        self.stream.write(")\n")
        self.stream.write("</span>\n")

    def visitBitand(self, node):
        self.stream.write("<span class='bitand'>\n")
        self.dispatch(node.nodes[0])
        for op in node._ops:
            self.stream.write("<span class='op'>\n")
            self.stream.write(self._text(op.name))
            self._popup(
                self._op(op)
                )
            self.stream.write("</span>\n")
            self.dispatch(op.expr)
        self.stream.write("</span>")

    def visitCallFunc(self, node):
        target = node._node
        targets = target.active()
        self.stream.write("<span class='callfunc'>\n")
        self.dispatch(node.node)
        self.stream.write("<span class='call'>\n")
        self.stream.write("(")
        self._popup(
            self._invocations(targets)
            )
        self.stream.write("</span>\n")
        first = 1
        for arg in node.args:
            if not first:
                self.stream.write(",\n")
            self.dispatch(arg)
            first = 0
        if node.star_args is not None:
            if not first:
                self.stream.write(", *\n")
            self.dispatch(node.star_args)
            first = 0
        if node.dstar_args is not None:
            if not first:
                self.stream.write(", **\n")
            self.dispatch(node.dstar_args)
            first = 0
        self.stream.write(")\n")
        self.stream.write("</span>\n")

    def visitCompare(self, node):
        self.stream.write("<span class='compare'>\n")
        self.dispatch(node.expr)
        for op in node._ops:
            self.stream.write("<span class='op'>\n")
            self.stream.write(self._text(op.name))
            self._popup(
                self._op(op)
                )
            self.stream.write("</span>\n")
            self.dispatch(op.expr)
        self.stream.write("</span>\n")

    def visitConst(self, node):
        if isinstance(node.value, (str, unicode)):
            self.stream.write("<span class='str'>\n")
        self.stream.write(repr(node.value))
        if isinstance(node.value, (str, unicode)):
            self.stream.write("</span>\n")

    def visitDict(self, node):
        self.stream.write("<span class='dict'>\n")
        self.stream.write("{")
        self._mapping(node)
        self.stream.write("}\n")
        self.stream.write("</span>\n")

    def visitDiv(self, node):
        self._visitBinary(node, "div", "/")

    def visitFloorDiv(self, node):
        self._visitBinary(node, "floordiv", "//")

    def visitGetattr(self, node):
        target = node._node
        targets = target.active()
        self.stream.write("<span class='getattr'>\n")
        self.dispatch(node.expr)
        self.stream.write("<span class='attr'>\n")
        self.stream.write(".")
        types = self._types(targets)
        if not target.is_annotated() or types:
            self._name_start(node.attrname)
        else:
            self._name_start(node.attrname, "no-types")
        self._popup(
            self._scopes(targets) +
            types
            )
        self._name_end()
        self.stream.write("</span>\n")
        self.stream.write("</span>\n")

    def visitKeyword(self, node):
        self.stream.write("<span class='keyword-arg'>\n")
        self.stream.write(node.name)
        self.stream.write("=")
        self.dispatch(node.expr)
        self.stream.write("</span>\n")

    def visitLambda(self, node):
        definition = node._node
        definitions = [n for n in definition.active() if not isinstance(n, Subprogram)]
        subprogram = node._subprogram
        subprograms = subprogram.active()
        self.stream.write("<span class='lambda'>\n")
        self._keyword("lambda")
        self._parameters(subprogram, subprograms)
        self.dispatch(node.code)
        self.stream.write("</span>\n")

    visitList = visitAssList

    def visitListComp(self, node):
        self.stream.write("<span class='listcomp'>\n")
        self.stream.write("[")
        self.dispatch(node.expr)
        for qual in node.quals:
            self.dispatch(qual)
        self.stream.write("]\n")
        self.stream.write("</span>\n")

    def visitListCompFor(self, node):
        self.stream.write("<span class='listcompfor'>\n")
        self.stream.write("<span class='iterator'>\n")
        self._keyword("for")
        self._popup(
            self._invocations(node._next_call.active())
            )
        self.stream.write("</span>\n")
        self.dispatch(node.assign)
        self.stream.write("<span class='iterator'>\n")
        self._keyword("in")
        self._popup(
            self._invocations(node._iter_call.active())
            )
        self.stream.write("</span>\n")
        self.dispatch(node.list)
        for if_ in node.ifs:
            self.dispatch(if_)
        self.stream.write("</span>\n")

    def visitListCompIf(self, node):
        conditional = node._node
        conditionals = conditional.active()
        self.stream.write("<span class='listcompif'>\n")
        self.stream.write("<span class='conditional'>\n")
        self._keyword("if")
        self._popup(
            self._invocations([c.test for c in conditionals])
            )
        self.stream.write("</span>\n")
        self.dispatch(node.test)
        self.stream.write("</span>\n")

    def visitMod(self, node):
        self._visitBinary(node, "mod", "%")

    def visitMul(self, node):
        self._visitBinary(node, "mul", "*")

    def visitName(self, node):
        target = node._node
        targets = target.active()
        types = self._types(targets)
        if not target.is_annotated() or types:
            self._name_start(target.name)
        else:
            self._name_start(target.name, "no-types")
        self._popup(
            self._scopes(targets) +
            types
            )
        self._name_end()

    def visitNot(self, node):
        self.stream.write("<span class='not'>\n")
        self._keyword("not")
        self.dispatch(node.expr)
        self.stream.write("</span>")

    def visitOr(self, node):
        self.stream.write("<span class='or'>\n")
        first = 1
        for n in node.nodes:
            if not first:
                self._keyword("or")
            self.dispatch(n)
            first = 0
        self.stream.write("</span>")

    def visitPower(self, node):
        self._visitBinary(node, "power", "**")

    def visitSlice(self, node):
        target = node._node
        targets = target.active()
        self.stream.write("<span class='slice'>\n")
        self.dispatch(node.expr)
        self.stream.write("<span class='call'>\n")
        self.stream.write("[")
        self._popup(
            self._invocations(targets)
            )
        self.stream.write("</span>\n")
        if node.lower:
            self.dispatch(node.lower)
        self.stream.write(":")
        if node.upper:
            self.dispatch(node.upper)
        # NOTE: Step?
        self.stream.write("]")
        self.stream.write("</span>\n")

    def visitSliceobj(self, node):
        self.stream.write("<span class='sliceobj'>\n")
        first = 1
        for n in node.nodes:
            if not first:
                self.stream.write(":")
            self.dispatch(n)
        self.stream.write("</span>\n")

    def visitSub(self, node):
        self._visitBinary(node, "sub", "-")

    def visitSubscript(self, node):
        target = node._node
        targets = target.active()
        self.stream.write("<span class='subscript'>\n")
        self.dispatch(node.expr)
        self.stream.write("<span class='call'>\n")
        self.stream.write("[")
        self._popup(
            self._invocations(targets)
            )
        self.stream.write("</span>\n")
        first = 1
        for sub in node.subs:
            if not first:
                self.stream.write(", ")
            self.dispatch(sub)
            first = 0
        self.stream.write("]")
        self.stream.write("</span>\n")

    visitTuple = visitAssTuple

    def visitUnaryAdd(self, node):
        self._visitUnary(node, "add", "+")

    def visitUnarySub(self, node):
        self._visitUnary(node, "sub", "-")

    # Output preparation methods.

    def _sequence(self, node):
        first = 1
        for n in node.nodes:
            if not first:
                self.stream.write(",\n")
            self.dispatch(n)
            first = 0

    def _mapping(self, node):
        first = 1
        for k, v in node.items:
            if not first:
                self.stream.write(",\n")
            self.dispatch(k)
            self.stream.write(":\n")
            self.dispatch(v)
            first = 0

    def _parameters(self, subprogram, subprograms):

        # Get all the parameter lists.

        params = []
        nparams = 0
        for sub in subprograms:
            params.append(sub.params)
            nparams = max(nparams, len(sub.params))
        stars = []
        have_star = 0
        for sub in subprograms:
            stars.append(sub.star)
            if sub.star is not None:
                have_star = 1
        dstars = []
        have_dstar = 0
        for sub in subprograms:
            dstars.append(sub.dstar)
            if sub.dstar is not None:
                have_dstar = 1

        # Traverse the parameter lists, choosing a "column" at a time.

        first = 1
        for n in range(0, nparams):
            if not first:
                self.stream.write(",\n")
            main_param, main_default = subprogram.params[n]
            self._name_start(main_param)
            self._popup(
                self._parameter(subprograms, params, n)
                )
            self._name_end()
            self._default(main_default)
            first = 0

        if have_star:
            if not first:
                self.stream.write(", *\n")
            main_param, main_default = subprogram.star
            self._name_start(main_param)
            self._popup(
                self._parameter(subprograms, stars)
                )
            self._name_end()
            self._default(main_default)
            first = 0

        if have_dstar:
            if not first:
                self.stream.write(", **\n")
            main_param, main_default = subprogram.dstar
            self._name_start(main_param)
            self._popup(
                self._parameter(subprograms, dstars)
                )
            self._name_end()
            self._default(main_default)
            first = 0

    def _parameter(self, subprograms, params, n=None):
        types = set()
        for i in range(0, len(subprograms)):
            subprogram = subprograms[i]
            if n is not None:
                param, default = params[i][n]
            else:
                param, default = params[i]
            if hasattr(subprogram, "paramtypes"):
                types.update(subprogram.paramtypes[param])
        return self._types_container(types, "types")

    def _default(self, default):
        if default is not None and default.original is not None:
            self.stream.write("=\n")
            self.dispatch(default.original)

    def _name_start(self, name, classes=None):
        if classes is not None:
            classes = " " + classes
        else:
            classes = ""
        self.stream.write("<span class='name%s'>%s\n" % (classes, name))

    def _name_end(self):
        self.stream.write("</span>\n")

    def _popup(self, info):
        if info:
            self.stream.write("<span class='popup'>\n")
            for section, subsection, labels in info:
                self.stream.write("<div class='%s'>\n" % section)
                for label in labels:
                    self.stream.write("<div class='%s'>\n" % subsection)
                    self.stream.write(label)
                    self.stream.write("</div>\n")
                self.stream.write("</div>\n")
            self.stream.write("</span>\n")

    def _op(self, node):
        if hasattr(node, "_left_call") and hasattr(node, "_right_call"):
            return self._invocations(node._left_call.active() + node._right_call.active())
        else:
            _node = node._node
            if isinstance(_node, Not):
                _node = _node.expr
            return self._invocations(_node.active())

    def _invocations(self, nodes):
        invocations = []
        for node in nodes:
            if hasattr(node, "invocations"):
                invocations += node.invocations

        # Record each link, avoiding duplicates.

        links = {}
        for invocation in invocations:
            fn = getattr(invocation, "copy_of", invocation).full_name()
            module = invocation.module.name
            name = invocation.name
            structures = [x.name for x in invocation.structures]
            qualified_name = ".".join([module] + structures + [name])

            # Record the label and the link texts.

            label = self._text(qualified_name)
            link = (self._url(module), self._url(fn))
            links[label] = link

        # Produce the list.

        if links:
            popup_labels = []
            for label, (module_name, target_name) in links.items():
                popup_labels.append("<a href='%s%sxhtml#%s'>%s</a>" % (module_name, os.path.extsep, target_name, label))
        else:
            popup_labels = []

        if popup_labels:
            return [("invocations", "invocation", popup_labels)]
        else:
            return []

    def _types(self, nodes):
        all_types = [(getattr(n, "types", []) or flatten(getattr(n, "writes", {}).values())) for n in nodes]
        types = flatten(all_types)
        return self._types_container(types, "types")

    def _types_container(self, types, style_class):
        labels = {}
        for type in types:
            fn = type.type.full_name()
            labels[self._text(fn)] = None

        if labels:
            return [(style_class, 'type', labels.keys())]
        else:
            return []

    def _raises(self, nodes):

        "Output the exception information for the given simplified 'nodes'."

        raises = set()
        for node in nodes:
            if hasattr(node, "raises") and node.raises:
                raises.update(node.raises)
        return self._types_container(raises, "raises")

    def _scopes(self, nodes):

        "Output the scope information for the given simplified 'nodes'."

        labels = {}
        for node in nodes:

            # Straightforward name loading/storing involves the local scope.

            if isinstance(node, StoreName) or isinstance(node, LoadName):
                labels["(local)"] = None

            # Other loading/storing involves attributes accessed on modules, classes
            # and objects.

            else:

                # Loading...

                if hasattr(node, "accesses") and node.accesses:
                    for ref, accesses in node.accesses.items():
                        fn = ref.full_name()
                        for attr, access in accesses:
                            access_fn = access.full_name()
                            label = self._text(fn)
                            if ref != access:
                                label += " (via " + self._text(access_fn) + ")"
                            labels[label] = None

                # Storing...

                if hasattr(node, "writes") and node.writes:
                    for ref in node.writes.keys():
                        fn = ref.full_name()
                        labels[self._text(fn)] = None

                # Non-loading...

                if hasattr(node, "non_accesses") and node.non_accesses:
                    self._types_container(node.non_accesses, "non-accesses")

                # Non-storing...

                if hasattr(node, "non_writes") and node.non_writes:
                    self._types_container(node.non_writes, "non-writes")

        if labels:
            return [("scopes", "scope", labels.keys())]
        else:
            return []

# Utility functions.

def flatten(lists):
    result = set()
    for l in lists:
        result.update(l)
    return result

# Convenience functions.

def browse(module, stream=None):
    browser = Browser(stream or sys.stdout)
    browser.process(module.original)

def makesummary(module, filename):
    stream = open(filename, "wb")
    try:
        summariser = Summariser(stream)
        summariser.process(module)
    finally:
        stream.close()

def makedoc(module, filename):
    stream = open(filename, "wb")
    try:
        browser = Browser(stream)
        browser.process(module.original)
    finally:
        stream.close()

def makedocs(module, modules, builtins):
    dirname = "%s-docs" % module.name
    if not os.path.exists(dirname):
        os.mkdir(dirname)
    for m in [module, builtins] + modules:
        makedoc(m, os.path.join(dirname, "%s%sxhtml" % (m.name, os.path.extsep)))
        makesummary(m, os.path.join(dirname, "%s%s%sxhtml" % (m.name, "-summary", os.path.extsep)))

# vim: tabstop=4 expandtab shiftwidth=4
